/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

/* According to POSIX 1003.1-2001 */
#include <sys/select.h>

/* According to earlier standards */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "messagebus.h"
#include "util.h"
#include "ifcontrol.h"
#include "iwcontrol.h"
#include "hostapcontrol.h"
#include "tcpclient.h"
#include "tcpsslclient.h"
#include "tcpsslapclient.h"
#include "tcpkismetclient.h"
#include "tcpheartclient.h"
#include "configfile.h"

// Global message bus
MessageBus mesbus;

int main(int argc, char *argv[]) {
    // Infrastructure bits
    TcpHeartClient hbclient;
    StdoutMessageClient stdclient;
    TcpSSLApClient sslcli;
    TcpSSLApMessageClient sslmsgcli;
    fd_set master_readset, master_writeset;
    ConfigFile *conf;
    Timetracker timetracker;

    // Other info
    int max_fd = 0;
    int ret = 0;
    char errstr[1024];
    unsigned int zone;

    // Build the message bus system, register the stdout client, link
    // the server into the message bus
    mesbus.RegisterClient(&stdclient, MSGFLAG_ALL);
    sslcli.RegisterMessageBus(&mesbus);

    
    // Link the SSL client into the sslmessage
    sslmsgcli.RegisterSSLClient(&sslcli);
    // Link into the msgbus
    mesbus.RegisterClient(&sslmsgcli, MSGFLAG_ALL);

    // Link into the hb client
    hbclient.RegisterTimetracker(&timetracker);
    hbclient.RegisterMessageBus(&mesbus);
    // Register the heartbeat tick once a second
    timetracker.RegisterTimer(SERVER_TIMESLICES_SEC,
                              NULL, 1, &HeartbeatClientTimer, 
                              (void *) &hbclient);

    // Link into the timer
    sslcli.RegisterTimetracker(&timetracker);

    conf = new ConfigFile;
    if (conf->ParseConfig("/usr/local/hope/hope_ap.conf") < 0) {
        fprintf(stderr, "FATAL: Failed to parse config file.\n");
        exit(1);
    }

    if (conf->FetchOpt("zone") != "") {
        if (sscanf(conf->FetchOpt("zone").c_str(), "%u", &zone) != 1) {
            snprintf(errstr, 1024, "Illegal value '%s' for zone in config file",
                     conf->FetchOpt("zone").c_str());
            mesbus.InjectMessage(errstr, MSGFLAG_ERROR | MSGFLAG_FATAL);
            exit(1);
        }
    } else {
        snprintf(errstr, 1024, "No zone value given in config file");
        mesbus.InjectMessage(errstr, MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    // Register the interfaces
    if (sslcli.RegisterEnvironment(conf->FetchOpt("name").c_str(),
                                   conf->FetchOpt("interface0").c_str(),
                                   conf->FetchOpt("interface1").c_str(),
                                   conf->FetchOpt("interface2").c_str(),
                                   conf->FetchOpt("interface3").c_str(),
                                   zone, conf->FetchOpt("kismetpath").c_str(),
                                   conf->FetchOpt("smitepath").c_str(),
                                   conf->FetchOpt("aphelperpath").c_str()) < 0) {
        mesbus.InjectMessage("Failed to register environment",
                             MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
   
    // ---- Connect to SSL server and start main loop
    
    // Set the server cert
    if (sslcli.SetCTXCert(conf->FetchOpt("certpem").c_str(), 
                          conf->FetchOpt("certpw").c_str()) < 0) {
        exit(1);
    }

    // Set the peer verification path
    if (sslcli.SetCTXVerify(conf->FetchOpt("rootpem").c_str(), 
                            conf->FetchOpt("rootpath").c_str()) < 0) {
        exit(1);
    }

    // Set the cipher list
    if (sslcli.SetCTXCipherList(conf->FetchOpt("cipherlist").c_str()) < 0) {
        exit(1);
    }

    string hbhostname;
    int hbport;

    if (ConfigFile::ParseHostLine(conf->FetchOpt("watchdog"), &hbhostname, 
                                  &hbport) < 0) {
        fprintf(stderr, "FATAL: Invalid watchdog host line '%s'\n", 
                conf->FetchOpt("watchdog").c_str());
        exit(1);
    }
  
    // Set up the server and start it
    for (unsigned int x = 0; x < 10; x++) {
        if (hbclient.Connect(hbhostname.c_str(), hbport) >= 0) 
            break;
        sleep(1);
    }

    if (hbclient.Valid() == 0) {
        fprintf(stderr, "FATAL: Didn't connect to the heartbeat server in "
                "10 seconds.\n");
        exit(1);
    }

    string hostname;
    int port;

    if (ConfigFile::ParseHostLine(conf->FetchOpt("apserver"), &hostname, &port) < 0) {
        fprintf(stderr, "FATAL: Invalid host line '%s'\n", 
                conf->FetchOpt("apserver").c_str());
        exit(1);
    }
  
    // Set up the server and start it
    for (unsigned int x = 0; x < 10; x++) {
        if (sslcli.Connect(hostname.c_str(), port) >= 0) 
            break;
        sleep(1);
    }

    if (sslcli.Valid() == 0) {
        fprintf(stderr, "FATAL: Didn't connect to the AP controller in 10 seconds.\n");
        exit(1);
    }

    // Free up the parser memory
    delete conf;
    conf = NULL;
   
    while (1) {
        struct timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = 100000;

        FD_ZERO(&master_readset);
        FD_ZERO(&master_writeset);

        // Get the sets from the central coordinator
        if ((max_fd = sslcli.MergeSet(master_readset, master_writeset,
                                      max_fd, &master_readset, &master_writeset)) < 0) {
            mesbus.InjectMessage(sslcli.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - mergeset failed\n");
            exit(1);
        }

        if ((max_fd = hbclient.MergeSet(master_readset, master_writeset,
                                        max_fd, &master_readset, 
                                        &master_writeset)) < 0) {
            mesbus.InjectMessage(hbclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - mergeset failed\n");
            exit(1);
        }

        if (select(max_fd + 1, &master_readset, &master_writeset, NULL, &tv) < 0) {
            mesbus.InjectMessage("select() failed", MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - select failed\n");
            exit(1);
        }

        if ((ret = sslcli.Poll(master_readset, master_writeset)) < 0) {
            mesbus.InjectMessage(sslcli.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - ssl poll failed\n");
            exit(1);
        } 

        if (sslcli.Valid() == 0) {
            printf("debug - ssl client failed\n");
            exit(1);
        }

        if ((ret = hbclient.Poll(master_readset, master_writeset)) < 0) {
            mesbus.InjectMessage(hbclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - heartbeat poll failed\n");
            exit(1);
        } 

        if (hbclient.Valid() == 0) {
            printf("debug - heartbeat client failed\n");
            exit(1);
        }

        timetracker.Tick();

    }

}

