/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "tcpclient.h"

TcpClient::TcpClient() {
    cl_valid = 0;
    remote_host[0] = '\0';
    remote_port = 0;
    client_fd = 0;
    connect_nonblocking = 1;
}

TcpClient::~TcpClient() {
    KillConnection();
}

int TcpClient::Connect(const char *in_remotehost, short int in_port) {
    snprintf(remote_host, MAXHOSTNAMELEN, "%s", in_remotehost);
    remote_port = in_port;

    if ((client_host = gethostbyname(in_remotehost)) == NULL) {
        snprintf(errstr, 1024, "TCP client could not resolve host \"%s\"",
                 in_remotehost);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    memset(&client_sock, 0, sizeof(client_sock));
    client_sock.sin_family = client_host->h_addrtype;
    memcpy((char *) &client_sock.sin_addr.s_addr, client_host->h_addr_list[0],
           client_host->h_length);
    client_sock.sin_port = htons(in_port);

    if ((client_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        snprintf(errstr, 1024, "TCP client socket() call failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Bind local half
    memset(&local_sock, 0, sizeof(local_sock));
    local_sock.sin_family = AF_INET;
    local_sock.sin_addr.s_addr = htonl(INADDR_ANY);
    local_sock.sin_port = htons(0);

    if (bind(client_fd, (struct sockaddr *) &local_sock, sizeof(local_sock)) < 0) {
        snprintf(errstr, 1024, "TCP client bind() failed: %s", strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Connect the sockets
    if (connect(client_fd, (struct sockaddr *) &client_sock, sizeof(client_sock)) < 0) {
        snprintf(errstr, 1024, "TCP client connect() failed: %s", strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Set nonblocking
    if (connect_nonblocking) {
        if (SetNonblocking() < 0)
            return -1;
    }
    
    cl_valid = 1;

    read_buf = new RingBuffer(CLI_RING_LEN);
    write_buf = new RingBuffer(CLI_RING_LEN);

    snprintf(errstr, 1024, "TcpClient connected to %s:%hd", in_remotehost, in_port);
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);

    return 1;
}

int TcpClient::SetNonblocking() {
    int save_mode = fcntl(client_fd, F_GETFL, 0);
    if (fcntl(client_fd, F_SETFL, save_mode | O_NONBLOCK) < 0) {
        snprintf(errstr, 1024, "TCP client failed to set nonblock fcntl: %s", 
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    return 0;
}

int TcpClient::KillConnection() {
    if (cl_valid) {
        delete read_buf;
        delete write_buf;
    }

    if (client_fd)
        close(client_fd);

    cl_valid = 0;

    return 1;
}

unsigned int TcpClient::MergeSet(fd_set in_rset, fd_set in_wset, 
                                 unsigned int in_max_fd,
                                 fd_set *out_rset, fd_set *out_wset) {
    unsigned int max;

    /*
    Trust the caller to have given us good sets
    FD_ZERO(out_rset);
    FD_ZERO(out_wset);
    */

    if ((int) in_max_fd < client_fd && cl_valid)
        max = client_fd;
    else
        max = in_max_fd;

    for (unsigned int x = 0; x <= max; x++) {
        if (FD_ISSET(x, &in_rset))
            FD_SET(x, out_rset);
        if (FD_ISSET(x, &in_wset))
            FD_SET(x, out_wset);
    }

    if (cl_valid) {
        FD_SET(client_fd, out_rset);

        if (write_buf->FetchLen() > 0)
            FD_SET(client_fd, out_wset);
    }

    return max;
}

int TcpClient::Poll(fd_set& in_rset, fd_set& in_wset) {
    int ret = 0;

    if (!cl_valid)
        return 0;

    // Look for stuff to read
    if (FD_ISSET(client_fd, &in_rset)) {
        // If we failed reading, die.  The read handler will kill the link.
        if ((ret = ReadBytes()) < 0)
            return ret;

        // If we've got new data, try to parse.  if we fail, die.
        if (ret != 0 && ParseData() < 0) {
            KillConnection();
            return -1;
        }
    }

    // Look for stuff to write
    if (FD_ISSET(client_fd, &in_wset)) {
        // If we can't write data, die.  The write handler will kill the link.
        if ((ret = WriteBytes()) < 0)
            return ret;
    }
    
    
    return ret;
}

char *TcpClient::FetchRemoteHost() {
    return remote_host;
}

short int TcpClient::FetchRemotePort() {
    return remote_port;
}

int TcpClient::ReadBytes() {
    uint8_t recv_bytes[1024];
    int ret;

    if ((ret = read(client_fd, recv_bytes, 1024)) < 0) {
        snprintf(errstr, 1024, "TCP client fd %d read() error: %s", 
                 client_fd, strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (ret == 0) {
        snprintf(errstr, 1024, "TCP client fd %d socket closed.", client_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (read_buf->InsertData(recv_bytes, ret) == 0) {
        snprintf(errstr, 1024, "TCP client fd %d read error, ring buffer full",
                 client_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    return ret;
}

int TcpClient::ParseData() {
    snprintf(errstr, 1024, "TCP client: No data parser.  How did this get called?");
    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
    return -1;
}

int TcpClient::WriteBytes() {
    uint8_t dptr[1024];
    int dlen, ret;

    write_buf->FetchPtr(dptr, 1024, &dlen);

    if ((ret = write(client_fd, dptr, dlen)) <= 0) {
        snprintf(errstr, 1024, "TCP client: Killing client fd %d write error %s",
                 client_fd, strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        KillConnection();
        return -1;
    }

    write_buf->MarkRead(ret);

    return ret;
}


