/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "endian_magic.h"
#include "tcpmonitorclient.h"

TcpMonitorClient::TcpMonitorClient() {
    tx_seqid = rand() % 32000;
    rx_seqid = 0;

    start_time = 0;
    
    config_delay = reconfig_delay = deadidle_delay = smite_delay = 
        smite_ignore = sniffers_per = rawtx_per = 0;

    zone_dirty = 0;
    ap_dirty = 0;
    if_dirty = 0;
}

TcpMonitorClient::~TcpMonitorClient() {

}

void TcpMonitorClient::FetchTimeLimits(unsigned int *ret_conf, unsigned int *ret_reconf,
                                       unsigned int *ret_dead, unsigned int *ret_smite,
                                       unsigned int *ret_smiteignore) {
    (*ret_conf) = config_delay;
    (*ret_reconf) = reconfig_delay;
    (*ret_dead) = deadidle_delay;
    (*ret_smite) = smite_delay;
    (*ret_smiteignore) = smite_ignore;
}

void TcpMonitorClient::FetchInterfaceAllocation(unsigned int *ret_sniffers, 
                                                unsigned int *ret_rawtx,
                                                mac_addr *ret_basebssid,
                                                char *ret_basessid) {
    (*ret_sniffers) = sniffers_per;
    (*ret_rawtx) = rawtx_per;
    (*ret_basebssid) = base_bssid;
    strncpy(ret_basessid, base_ssid, 32);
}

int TcpMonitorClient::ParseData() {
    while (read_buf->FetchLen() >= (int) MONLEN_HEADER) {
        monitor_frame fmon;
        int flen;

        read_buf->FetchPtr((uint8_t *) &fmon, MONLEN_HEADER, &flen);

        if (fmon.sentinel != htonl(MONITOR_SENTINEL)) {
            snprintf(errstr, 1024, "TCPMonitorClient got invalid monitor frame sentinel, "
                     "closing connection");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }

        fmon.len = (uint32_t) ntohl(fmon.len);

        // Drop out of the loop if we're a fragment
        if (read_buf->FetchLen() < (int) (MONLEN_HEADER + fmon.len))
            break;

        // Fetch it into the offset
        read_buf->MarkRead(MONLEN_HEADER);
        read_buf->FetchPtr(((uint8_t *) &fmon) + MONLEN_HEADER, fmon.len, &flen);
        read_buf->MarkRead(fmon.len);

        // Check rxseqid
        long long unsigned int fseq = kis_ntoh64(fmon.sequence_id);
        if ((fseq <= rx_seqid || fseq >= rx_seqid + MONITOR_SEQ_WINDOW) && rx_seqid != 0) {
            snprintf(errstr, 1024, "TCPMonitorClient got invalid sequence ID, got %llu "
                     "but expected >= %llu", fseq, (long long unsigned int) rx_seqid);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }
        rx_seqid = fseq;

        // Handle the frame
        if (fmon.type == MONTYPE_SERVERINFO) {
            if (fmon.u.serverinfo.proto_ver != MONITOR_PROTOVER) {
                snprintf(errstr, 1024, "TCPMonitorClient got invalid monitor version %d "
                         "expected %d, terminating connection.", fmon.u.serverinfo.proto_ver,
                         MONITOR_PROTOVER);
                return -1;
            }

            start_time = (time_t) ntohl(fmon.u.serverinfo.start_time);
            config_delay = ntohs(fmon.u.serverinfo.config_delay);
            reconfig_delay = ntohs(fmon.u.serverinfo.reconfig_delay);
            deadidle_delay = ntohs(fmon.u.serverinfo.deadidle_delay);
            smite_delay = ntohs(fmon.u.serverinfo.smite_delay);
            smite_ignore = ntohs(fmon.u.serverinfo.smite_ignore);
            sniffers_per = ntohs(fmon.u.serverinfo.sniffers_per);
            rawtx_per = ntohs(fmon.u.serverinfo.rawtx_per);

            base_bssid = mac_addr(fmon.u.serverinfo.base_bssid, 
                                  fmon.u.serverinfo.base_bssidmask);

            strncpy(base_ssid, (char *) fmon.u.serverinfo.base_ssid, 32);
        } else if (fmon.type == MONTYPE_ZONE) {
            zone *zn;

            unsigned int zid = ntohs(fmon.u.zone.id);
            if (zone_id_map.find(zid) != zone_id_map.end()) {
                zn = zone_id_map[zid];
            } else {
                zn = new zone;
                zone_id_map[zid] = zn;
            }

            zn->id = zid;
            zn->dirty = fmon.u.zone.dirty;
            zn->num_sniffers = fmon.u.zone.nsniffer;
            zn->num_rawtx = fmon.u.zone.nrawtx;
            zn->num_ap = fmon.u.zone.nap;
            zn->num_nodes = fmon.u.zone.nnodes;
            zn->num_interfaces = fmon.u.zone.ninterfaces;
            zn->num_alerts = ntohl(fmon.u.zone.nalerts);
            zn->num_attacks = ntohl(fmon.u.zone.nattacks);
            zn->update_time = time(0);

            zone_dirty = 1;
        } else if (fmon.type == MONTYPE_AP) {
            ap *an;

            unsigned int sid = ntohs(fmon.u.ap.stationid);
            if (ap_id_map.find(sid) != ap_id_map.end()) {
                an = ap_id_map[sid];
            } else {
                an = new ap;
                ap_id_map[sid] = an;
            }

            strncpy(an->name, (char *) fmon.u.ap.name, 32);
            an->zoneid = ntohs(fmon.u.ap.zoneid);
            an->stationid = sid;
            an->configured = fmon.u.ap.configured;
            an->active = fmon.u.ap.active;
            an->dead = fmon.u.ap.dead;
            an->last_helo = ntohl(fmon.u.ap.last_helo);
            an->num_interfaces = fmon.u.ap.ninterfaces;
            an->loadavg_major = fmon.u.ap.loadavg_major;
            an->loadavg_minor = fmon.u.ap.loadavg_minor;
            an->update_time = time(0);

            ap_dirty = 1;
        } else if (fmon.type == MONTYPE_INTERFACE) {
            interface *ifc;

            unsigned int ifid = ((ntohs(fmon.u.interface.stationid) << 16) | fmon.u.interface.index);
            if (if_id_map.find(ifid) != if_id_map.end()) {
                ifc = if_id_map[ifid];
            } else {
                ifc = new interface;
                if_id_map[ifid] = ifc;
            }

            strncpy(ifc->name, (char *) fmon.u.interface.name, 32);
            ifc->stationid = ntohs(fmon.u.interface.stationid);
            ifc->index = fmon.u.interface.index;
            ifc->mac = mac_addr(fmon.u.interface.mac);
            ifc->innate_mac = mac_addr(fmon.u.interface.innate_mac);
            ifc->channel = ntohs(fmon.u.interface.channel);
            ifc->cf_bitmap = ntohl(fmon.u.interface.cf_bitmap);
            ifc->last_mac_change = ntohl(fmon.u.interface.last_mac_change);
            ifc->num_clients = ntohl(fmon.u.interface.nclients);
            ifc->framecount = kis_ntoh64(fmon.u.interface.framecount);
            ifc->errorcount = kis_ntoh64(fmon.u.interface.errorcount);

            ifc->tx_packetcount = kis_ntoh64(fmon.u.interface.txpacketcount);
            ifc->rx_packetcount = kis_ntoh64(fmon.u.interface.rxpacketcount);
            ifc->tx_bytecount = kis_ntoh64(fmon.u.interface.txbytecount);
            ifc->rx_bytecount = kis_ntoh64(fmon.u.interface.rxbytecount);
           
            ifc->update_time = time(0);
           
            if_dirty = 1; 
        } else if (fmon.type == MONTYPE_MESSAGE) {
            int len = ntohl(fmon.u.message.msglen);

            if (len > 0 && len < 1024) {
                fmon.u.message.msgtext[len] = 0;
                snprintf(errstr, 1024, "TCPMonitorClient - %s", (char *) fmon.u.message.msgtext);
                msgbus->InjectMessage(errstr, ntohl(fmon.u.message.msgflags));
            } else {
                snprintf(errstr, 1024, "TCPMonitorClient - Corrupt message frame, skipping.");
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            }
        } else {
            snprintf(errstr, 1024, "TCPMonitorClient got unknown frame type %d, skipping.",
                     fmon.type);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        }
    }

    return 1;
}

