/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TCPSSLAPCLIENT_H__
#define __TCPSSLAPCLIENT_H__

#include "tcpsslclient.h"
#include "tcpkismetclient.h"
#include "ap_stream_proto.h"
#include "macaddr.h"
#include "hostapcontrol.h"

// Time tick to update the status and harvest external commands
int SSLApClientUpdateStatsTimer(Timetracker::timer_event *evt, void *parm);
// Time tick to send a helo frame
int SSLApClientSendHeloTimer(Timetracker::timer_event *evt, void *parm);
// Interface thrash handler
int SSLApClientInterfaceThrashTimer(Timetracker::timer_event *evt, void *parm);

class TcpSSLApClient : public TcpSSLClient {
public:
    TcpSSLApClient();

    // Handle sending the announce frame
    virtual int Connect(const char *in_remotehost, short int in_port);

    // Register interfaces and anme
    virtual int RegisterEnvironment(const char *in_name,
                                    const char *in_if0,
                                    const char *in_if1,
                                    const char *in_if2,
                                    const char *in_if3,
                                    unsigned int in_zone,
                                    const char *in_kispath,
                                    const char *in_smitepath,
                                    const char *in_aphelperpath);

    // Send a message
    virtual int SendMessagePacket(const char *in_msg, int in_flags);

    // Override some other stuff so we can merge it with the kismet clients 
    // on each interface
    virtual unsigned int MergeSet(fd_set in_rset, fd_set in_wset, 
                                  unsigned int in_max_fd,
                                  fd_set *out_rset, fd_set *out_wset);
    virtual int Poll(fd_set& in_rset, fd_set& in_wset);

protected:
    typedef enum ProtocolState {
        // No state
        PS_PENDING_REGISTRATION,
        // We're waiting to send the announce frame
        PS_PENDING_ANNOUNCE,
        // We're connected to the SSL but we're pending acceptance of our
        // announcement
        PS_PENDING_CONFIGURE, 
        // We're configured
        PS_CONFIGURED,
        // We're running normally
        PS_RUNNING
    };

    // Track local attributes
    typedef struct local_interface {
        local_interface() {
            memset(hwaddr, 0, 6);
            memset(innate_hwaddr, 0, 6);
            ifname[0] = '\0';
            kisclient = NULL;
            kismet_pid = 0;
            kismet_port = 0;
            kismet_time = 0;
            kismet_pending = 0;
            rawtx_pid = 0;
            cf_bitmap = 0;
            old_cf_bitmap = 0;
            channel = 0;
            old_channel = 0;
            clientcount = 0;
            framecount = 0;
            errorcount = 0;
            dirty = 0;
            macdirty = 0;
            failures = 0;
            last_kis_alert = 0;
            last_kis_network = 0;
            rx_packetcount = 0;
            tx_packetcount = 0;
            rx_bytecount = 0;
            tx_bytecount = 0;
        }

        // Interface name
        char ifname[32];
        // hwaddr
        uint8_t hwaddr[6];
        uint8_t innate_hwaddr[6];

        // Link to the kismet sniffer if we need to be running one
        TcpKismetClient *kisclient;
        // pid of the sniffer if we're running one
        pid_t kismet_pid;
        // Port kismet is on
        unsigned int kismet_port;
        // Time we last heard from kismet
        time_t kismet_time;
        // Number of times we've failed to connect
        int kismet_failedcon;
        // Pending a connect
        int kismet_pending;
        // Last alert timestamp we handled
        time_t last_kis_alert;
        // Last network timestamp we handled
        time_t last_kis_network;

        // pid of the rawtx engine if we're running one - we don't have an
        // interactive link per se
        pid_t rawtx_pid;
        // Spawn time of rawtx app
        time_t rawtx_time;

        // Config bitmap of what state the card is in
        uint32_t cf_bitmap;
        // Old bitmap of what we looked like before, if we're dirty
        uint32_t old_cf_bitmap;

        // Channel
        unsigned int channel;
        unsigned int old_channel;

        // Is the config dirty?
        int dirty;
        int macdirty;

        // Number of failures in a row
        int failures;
        
        // Interface counts
        unsigned int clientcount;
        unsigned long long int framecount;
        unsigned long long int errorcount;

        unsigned long long int rx_packetcount;
        unsigned long long int tx_packetcount;
        unsigned long long int rx_bytecount;
        unsigned long long int tx_bytecount;

        // These are never masked so a macmap isn't needed
        map<mac_addr, hostap_station_rec> station_map;
    };

    // Handle returning data
    virtual int ParseData();

    // Handle individual packet types
    virtual int ParseConfigure(mn_configure_packet *in_conf);
    virtual int ParseActivate(mn_activate_packet *in_act);
    virtual int ParseBSSID(mn_bssid_packet *in_bssid);
    virtual int ParseSmite(mn_smite_packet *in_smite);

    // Send a HELO -- all the data is contained internally
    virtual int SendHeloPacket();
    // Send an announce - all the data is, again, internal
    virtual int SendAnnouncePacket();
    // Send an alert packet
    virtual int SendKisAlertPacket(TcpKismetClient::alert_rec *in_alert);
    // Send info about a network
    virtual int SendKisNetworkPacket(TcpKismetClient::network_rec *in_net);

    // Insert a packet into the ringbuffer for writing
    virtual int QueuePacket(stream_frame *in_frame);

    // Update stats on interfaces
    virtual int UpdateStats();

    // Configure interfaces
    virtual int ActivateInterfaces();

    // Connect to a kismet instance
    virtual int ConnectKismetInterface(unsigned int in_ifnum);

    // Look for 'downer' interfaces and other thrash conditions
    virtual int ThrashTick();

    // Our current state
    ProtocolState state;

    // Timer events
    int stats_event_id;
    int helo_event_id;
    int thrash_event_id;

    // Kismet stuff - port allocation and paths
    unsigned int kisport;
    char *kispath;
    char *smitepath;
    char *aphelperpath;

    // Station ID
    unsigned int station_id;

    // SSID
    char ssid[32];

    // Physical zone
    unsigned int zone;

    // Name
    char station_name[32];

    // Interfaces
    local_interface interfaces[4];
    
    // Sequence ID
    unsigned long int tx_seqid;
    unsigned long int rx_seqid;

    // Load average
    uint8_t load_maj, load_min;

    friend int SSLApClientUpdateStatsTimer(Timetracker::timer_event *evt, void *parm);
    friend int SSLApClientSendHeloTimer(Timetracker::timer_event *evt, void *parm);
    friend int SSLApClientConnectKismetTimer(Timetracker::timer_event *evt, void *parm);
    friend int SSLApClientInterfaceThrashTimer(Timetracker::timer_event *evt, 
                                               void *parm);

};

// Kicker struct for the timer event
typedef struct TcpSSLApClientKismetTimerRec {
    TcpSSLApClient *cli;
    unsigned int ifnum;
};

int SSLApClientConnectKismetTimer(Timetracker::timer_event *evt, void *parm);

class TcpSSLApMessageClient : public MessageClient {
public:
    void ProcessMessage(string in_msg, int in_flags);
    void RegisterSSLClient(TcpSSLApClient *in_cli) {
        cli = in_cli;
    }
protected:
    TcpSSLApClient *cli;
};


#endif

