/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "tcpsslapserver.h"
#include "endian_magic.h"

int SSLApServerConfigureTimer(Timetracker::timer_event *evt, void *parm) {
    // Just kick off the function inside the server and don't repeat
    TcpSSLApServer *srv = (TcpSSLApServer *) parm;

    srv->ConfigureAllAps();

    return 0;
}

int SSLApServerActiveEventTimer(Timetracker::timer_event *evt, void *parm) {
    // Kick the class
    TcpSSLApServer *srv = (TcpSSLApServer *) parm;

    srv->ProcessActiveEvents();

    return 1;
}

TcpSSLApServer::TcpSSLApServer() {
    next_station_id = 1;
    config_delay = reconfigure_delay = deadidle_delay = 0;
    alerts_dirty = networks_dirty = 0;
    active_id = 0;
}

void TcpSSLApServer::KillConnection(int in_fd) {
    // printf("debug - killing connection %d\n", in_fd);
    map<int, TcpSSLApServer::apserver_aprec *>::iterator litr = 
        linked_ap_map.find(in_fd);

    if (litr != linked_ap_map.end()) {
        // Flag us as unconnected
        litr->second->dead = 1;
        litr->second->last_packet_local = time(0);
    }
    
    return TcpSSLServer::KillConnection(in_fd);
}

void TcpSSLApServer::SetTimeLimits(unsigned int in_conf, unsigned int in_reconf, 
                                   unsigned int in_idle, unsigned int in_smitedelay,
                                   unsigned int in_smiteignore) {
    config_delay = in_conf;
    reconfigure_delay = in_reconf;
    deadidle_delay = in_idle;
    smite_delay = in_smitedelay;
    smite_ignore = in_smiteignore;

    if (active_id == 0)
        active_id = timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * 5,
                                               NULL, 1, &SSLApServerActiveEventTimer, 
                                               (void *) this);
}

void TcpSSLApServer::SetInterfaceAllocation(unsigned int in_sniffers,
                                            unsigned int in_rawtx,
                                            mac_addr in_base_bssid,
                                            const char *in_base_ssid) {
    sniffers_per_zone = in_sniffers;
    rawtx_per_zone = in_rawtx;
    next_bssid = in_base_bssid;
    next_bssid++;
    strncpy(base_ssid, in_base_ssid, 32);
}

void TcpSSLApServer::Tick() {
    time_t curtime = time(0);

    // Do maintenance on all the APs in the list
    for (map<int, apserver_aprec *>::iterator aitr = linked_ap_map.begin();
         aitr != linked_ap_map.end(); ++aitr) {

        // Time out ones that aren't linked
        if (aitr->second->dead &&
            (curtime - aitr->second->last_packet_local > (int) reconfigure_delay)) {

            snprintf(errstr, 1024, "TCPSSLAP Didn't hear from dead "
                     "AP id %u '%s' for more than %us, dumping them.",
                     aitr->second->station_id, aitr->second->name, 
                     reconfigure_delay);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

            // Destroy them - we need to preserve the old iterator though
            // and reset to it.  bail out if theres nothing in our
            // map.
            map<int, apserver_aprec *>::iterator ditr = aitr;
            int beg = 0;

            if (aitr != linked_ap_map.begin())
                --ditr;
            else
                beg = 1;

            DestroyRecords(aitr->second);

            if (linked_ap_map.size() == 0)
                break;

            if (beg == 1)
                aitr = linked_ap_map.begin();
            else
                aitr = ditr;

            continue;
        }
        
        // Don't idle out inactive or unconfigured ones or dead ones
        if (aitr->second->configured == 0 || aitr->second->active == 0 ||
            aitr->second->dead)
            continue;

        // If someone stopped talking to us, boot them out
        if (curtime - aitr->second->last_packet_local > (int) deadidle_delay) {
            snprintf(errstr, 1024, "TCPSSLAP Didn't hear from AP id %u "
                     "name '%s' for more than %u seconds, marking dead.",
                     aitr->second->station_id, aitr->second->name, deadidle_delay);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            KillConnection(aitr->first);
            continue;
        }
    }
}

void TcpSSLApServer::ConfigureAllAps() {
    // printf("debug - config all aps\n");
    // This does the actual voodoo of counting interfaces and assigning them
    // to channels and other activities
    
    // Map of zone numbers to APs
    map<int, apserver_zonerec> build_zone_map;

    // Split the aps into zones and count interfaces
    for (map<int, apserver_aprec *>::iterator aitr = linked_ap_map.begin();
         aitr != linked_ap_map.end(); ++aitr) {
        // Don't mess with zones that aren't dirty...  Every possible ap
        // zone is in this map since when we accept them we put it in...
        if (zone_map[aitr->second->zone].zone_dirty == 0) {
            printf("debug - zone %u is clean, skipping interfaces in it.\n",
                   aitr->second->zone);
            continue;
        }

        build_zone_map[aitr->second->zone].aps.push_back(aitr->second);

        // Set the local config flag in anticipation of completing
        aitr->second->localconf = 1;
        aitr->second->configured = 0;
       
        // Learn the interfaces
        for (unsigned int cn = 0; cn < 4; cn++) {
            if ((aitr->second->interfaces[cn].cf_bitmap & AP_IFFLAG_AVAILABLE) &&
                (aitr->second->interfaces[cn].cf_bitmap & AP_IFFLAG_ERROR) == 0) {
                build_zone_map[aitr->second->zone].interfaces.push_back(&(aitr->second->interfaces[cn]));
                // Give the interface an iterative BSSID
                // Broken.  We set their mac when we learn the interface initially.
                // aitr->second->interfaces[cn].mac = next_bssid++;
                // Set the iterator
                aitr->second->interfaces[cn].conf_iter_set = 0;
                // Learn this mac
                bssid_if_map.insert(aitr->second->interfaces[cn].mac, 
                                    &(aitr->second->interfaces[cn]));
                printf("debug - learned interface %d assigned mac %s\n",
                       cn, aitr->second->interfaces[cn].mac.Mac2String().c_str());
            }
        }
    }

    // Look at each zone
    for (map<int, apserver_zonerec>::iterator zitr = build_zone_map.begin();
         zitr != build_zone_map.end(); ++zitr) {

        // Figure out how many we should allocate per...
        unsigned int ntotal = zitr->second.interfaces.size();
        unsigned int naps = ntotal;
        unsigned int nsniff = sniffers_per_zone;
        unsigned int nraw = rawtx_per_zone;

        // Figure out if we have space to do sniffing and rawtx...
        if (naps < 3) {
            snprintf(errstr, 1024, "TCPSSLAP Zone %u has less than 3 interfaces "
                     "active.  This is too few to monitor and respond, dropping "
                     "to pure bridge mode with no smart functionality.",
                     zitr->first);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

            unsigned int choffset = MAX_CHANNEL / naps;

            for (unsigned int ifn = 0; ifn < naps; ifn++) {
                printf("debug - assigned interface %d channel %d\n",
                       ifn, (ifn * choffset) + 1);
                zitr->second.interfaces[ifn]->conf_iter_set = 1;
                zitr->second.interfaces[ifn]->channel = (ifn * choffset) + 1;
                zitr->second.interfaces[ifn]->cf_bitmap = 
                    (AP_IFFLAG_AVAILABLE | AP_IFFLAG_ACCESSPOINT | AP_IFFLAG_CONFIGURED);
            }
        
            zone_map[zitr->first].zone_dirty = 0;
            zone_map[zitr->first].num_sniffer = 0;
            zone_map[zitr->first].num_rawtx = 0;
            zone_map[zitr->first].num_ap = naps;
            zone_map[zitr->first].aps = zitr->second.aps;
            zone_map[zitr->first].interfaces = zitr->second.interfaces;

            continue;
        }

        // Number of free interfaces to become APs
        naps = ntotal - nsniff - nraw;

        // Trim rawtx if we have too few APs
        if ((naps < nraw || naps <= 2) && nraw > 1) {
            snprintf(errstr, 1024, "TCPSSLAP Zone %u does not have enough active "
                     "interfaces to support %u raw transmitters, dropping to 1",
                     zitr->first, nraw);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            nraw = 1;
            naps = ntotal - nsniff - nraw;
        }

        // Trim sniffers if we have too few APs
        if ((naps < nsniff || naps <= 2) && nsniff > 1) {
            snprintf(errstr, 1024, "TCPSSLAP Zone %u does not have enough active "
                     "interfaces to support %u sniffers, dropping to 1",
                     zitr->first, nsniff);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            nsniff = 1;
            naps = ntotal - nsniff - nraw;
        }

        if (naps <= 2) {
            // Warn on too few AP interfaces
            snprintf(errstr, 1024, "TCPSSLAP Zone %u has %u interfaces in AP "
                     "mode.  This may lead to useability problems.",
                     zitr->first, naps);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        } else if (naps > (MAX_CHANNEL / 2)) {
            // Warn on too many AP interfaces
            snprintf(errstr, 1024, "TCPSSLAP Zone %u has %u AP interfaces.  This "
                     "oversaturates the spectrum and isn't the most efficient "
                     "use of hardware.", zitr->first, naps);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        }

        snprintf(errstr, 1024, "TCPSSLAP Zone %u has %u APs and %u interfaces "
                 "%u sniffer, %u rawtx, %u AP mode",
                 zitr->first, zitr->second.aps.size(), ntotal, nsniff, nraw, naps);
        msgbus->InjectMessage(errstr, MSGFLAG_INFO);

        // The interfaces are in here in more or less the order of where
        // they're located in the APs...  So don't assign 2 adjacent ones to be 
        // sniffers or the same channels and we should be ok...
            
        // Assign sniffers distributed throughout the interface list
        // First we try to assign them to interfaces that are already
        // sniffers - no reason to flap around and risk mode change problems.
        for (unsigned int nif = 0; nif < zitr->second.interfaces.size() &&
             nsniff > 0; nif++) {
            if ((zitr->second.interfaces[nif]->cf_bitmap & AP_IFFLAG_SNIFFER) &&
                zitr->second.interfaces[nif]->conf_iter_set == 0) {
                zitr->second.interfaces[nif]->conf_iter_set = 1;
                nsniff--;
            }
        }

        if (nsniff > 0) {
            unsigned int nsniffskip = ntotal / nsniff;
            for (unsigned int sn = 0; sn < nsniff; sn++) {
                // Find an unassigned interface - we're armored because it would have
                // adjusted the sniffer and rawtx counts earlier
                unsigned int pos = ((sn + 1) * nsniffskip) - 1;
                while (zitr->second.interfaces[pos]->conf_iter_set != 0) {
                    if (++pos >= ntotal - 1)
                        pos = 0;
                }

                // Assign it to a sniffer
                zitr->second.interfaces[pos]->cf_bitmap = 
                    (AP_IFFLAG_AVAILABLE | AP_IFFLAG_SNIFFER | AP_IFFLAG_CONFIGURED);
                zitr->second.interfaces[pos]->conf_iter_set = 1;
            }
        }

        // Assign rawtx distributed throughout the interface list
        // Try to assign to old rawtx units again too
        for (unsigned int nif = 0; nif < zitr->second.interfaces.size() &&
             nraw > 0; nif++) {
            if ((zitr->second.interfaces[nif]->cf_bitmap & AP_IFFLAG_RAWTX) &&
                zitr->second.interfaces[nif]->conf_iter_set == 0) {
                zitr->second.interfaces[nif]->conf_iter_set = 1;
                nraw--;
            }
        }

        if (nraw > 0) {
            unsigned int nrawskip = ntotal / nraw;
            for (unsigned int rn = 0; rn < nraw; rn++) {
                // Find an unassigned interface - we're armored because it would have
                // adjusted the sniffer and rawtx counts earlier
                unsigned int pos = ((rn + 1) * nrawskip) - 1;
                while (zitr->second.interfaces[pos]->conf_iter_set != 0) {
                    if (++pos >= ntotal - 1)
                        pos = 0;
                }

                // Assign it to a sniffer
                zitr->second.interfaces[pos]->cf_bitmap = 
                    (AP_IFFLAG_AVAILABLE | AP_IFFLAG_RAWTX | AP_IFFLAG_CONFIGURED);
                zitr->second.interfaces[pos]->conf_iter_set = 1;
            }
        }

        // Assign channels
        unsigned int nchanskip = MAX_CHANNEL / naps;
        // Assigned number
        unsigned int an = 0;
        for (unsigned int cn = 0; cn < ntotal; cn++) {
            if (zitr->second.interfaces[cn]->conf_iter_set != 0)
                continue;

            zitr->second.interfaces[cn]->cf_bitmap =
                (AP_IFFLAG_AVAILABLE | AP_IFFLAG_ACCESSPOINT | AP_IFFLAG_CONFIGURED);

            zitr->second.interfaces[cn]->channel = ((an * nchanskip) % MAX_CHANNEL) + 1;
            an++;

            printf("debug - assigned interface %u in zone %d to be ap on "
                   "channel %u, num %u assigned\n", cn, zitr->first,
                   zitr->second.interfaces[cn]->channel, an);
        }

        zone_map[zitr->first].zone_dirty = 0;
        zone_map[zitr->first].num_sniffer = nsniff;
        zone_map[zitr->first].num_rawtx = nraw;
        zone_map[zitr->first].num_ap = naps;
        zone_map[zitr->first].aps = zitr->second.aps;
        zone_map[zitr->first].interfaces = zitr->second.interfaces;
    }

    for (map<int, apserver_aprec *>::iterator aitr = linked_ap_map.begin();
         aitr != linked_ap_map.end(); ++aitr) {
        if (aitr->second->configured == 0 && aitr->second->dead == 0) {
            ConfigureAp(aitr->second);
            aitr->second->configured = 1;
            ActivateAp(aitr->second);
            aitr->second->active = 1;
        }
    }

    config_pending = 0;
}

void TcpSSLApServer::ConfigureAp(apserver_aprec *in_arec) {
    // printf("debug - configureap\n");
    if (!in_arec->localconf) {
        snprintf(errstr, 1024, "TCPSSLAP Unable to push config to "
                 "AP %u '%s', no local config to push.", 
                 in_arec->station_id, in_arec->name);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    // Can't push a config if we're dead - leave it to get pushed on 
    // reconnect.
    if (in_arec->dead)
        return;

    // Nerver tell an AP that its broken
    in_arec->interfaces[0].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    in_arec->interfaces[1].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    in_arec->interfaces[2].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    in_arec->interfaces[3].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);

    in_arec->last_packet_local = time(0);

    // Push a config down to the AP
    stream_frame retframe;
    mn_configure_packet confpay;

    retframe.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    retframe.frame_type = FRAMETYPE_CONFIGURE;
    retframe.frame_len = (uint32_t) htonl(FRAMELEN_CONFIGURE);
    retframe.payload = (uint8_t *) &confpay;

    confpay.station_id = (uint16_t) htons(in_arec->station_id);

    strncpy((char *) confpay.ssid, base_ssid, 32);

    confpay.ifchannel0 = (uint16_t) htons(in_arec->interfaces[0].channel);
    confpay.ifchannel1 = (uint16_t) htons(in_arec->interfaces[1].channel);
    confpay.ifchannel2 = (uint16_t) htons(in_arec->interfaces[2].channel);
    confpay.ifchannel3 = (uint16_t) htons(in_arec->interfaces[3].channel);

    confpay.cf_bitmap0 = (uint32_t) htonl(in_arec->interfaces[0].cf_bitmap);
    confpay.cf_bitmap1 = (uint32_t) htonl(in_arec->interfaces[1].cf_bitmap);
    confpay.cf_bitmap2 = (uint32_t) htonl(in_arec->interfaces[2].cf_bitmap);
    confpay.cf_bitmap3 = (uint32_t) htonl(in_arec->interfaces[3].cf_bitmap);

    // This is a shitty way to do this but it'll do for now
    for (unsigned int x = 0; x < 6; x++)
        confpay.bssid0[x] = in_arec->interfaces[0].mac[x];
    for (unsigned int x = 0; x < 6; x++)
        confpay.bssid1[x] = in_arec->interfaces[1].mac[x];
    for (unsigned int x = 0; x < 6; x++)
        confpay.bssid2[x] = in_arec->interfaces[2].mac[x];
    for (unsigned int x = 0; x < 6; x++)
        confpay.bssid3[x] = in_arec->interfaces[3].mac[x];

    // Update the last time we set the MAC on this interface
    for (unsigned int x = 0; x < 4; x++)
        in_arec->interfaces[x].last_mac_change = time(0);
    
    // Queue it out
    QueuePacket(in_arec->fd, in_arec, &retframe);
}

void TcpSSLApServer::PushBSSID(apserver_aprec *in_arec, int in_index) {
    // printf("debug - pushbssid\n");
    // Can't push a config if we're dead - leave it to get pushed on 
    // reconnect.
    if (in_arec->dead)
        return;

    // Push a config down to the AP
    stream_frame retframe;
    mn_bssid_packet bssidpay;

    retframe.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    retframe.frame_type = FRAMETYPE_CHBSSID;
    retframe.frame_len = (uint32_t) htonl(FRAMELEN_CHBSSID);
    retframe.payload = (uint8_t *) &bssidpay;

    bssidpay.if_index = (uint8_t) (in_index & 0xFF);

    for (unsigned int x = 0; x < 6; x++) 
        bssidpay.bssid[x] = in_arec->interfaces[in_index].mac[x];

    // Queue it out
    QueuePacket(in_arec->fd, in_arec, &retframe);
}

void TcpSSLApServer::ActivateAp(apserver_aprec *in_arec) {
    // printf("debug - activateap\n");
    stream_frame actframe;
    mn_activate_packet actpay;

    actframe.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    actframe.frame_type = FRAMETYPE_ACTIVATE;
    actframe.frame_len = (uint32_t) htonl(FRAMELEN_ACTIVATE);
    actframe.payload = (uint8_t *) &actpay;

    // Nothing to do in the payload other than the sequence id
    QueuePacket(in_arec->fd, in_arec, &actframe);
}

void TcpSSLApServer::DestroyRecords(apserver_aprec *in_arec) {
    // Mark this zone as dirty
    zone_map[in_arec->zone].zone_dirty = 1;

    // Find them in the linked fd map and kill them
    map<int, apserver_aprec *>::iterator litr = linked_ap_map.find(in_arec->fd);
    if (litr != linked_ap_map.end()) {
        linked_ap_map.erase(litr->first);
    }

    delete in_arec;
    in_arec = NULL;

    // Schedule a configure
    if (config_pending == 0) {
        timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * config_delay,
                                   NULL, 0, &SSLApServerConfigureTimer, 
                                   (void *) this);
        config_pending = 1;
    }
}

int TcpSSLApServer::ParseData(int in_fd) {
    // printf("debug - parsedata\n");
    RingBuffer *rbuf = read_buf_map[in_fd];
    int ret = 0;

    // Look for something long enough to hold a frame
    while (rbuf->FetchLen() >= (int) FRAMELEN_HEADER) {
        // Map it into the frame header
        stream_frame fhead;
        int flen;
    
        rbuf->FetchPtr((uint8_t *) &fhead, FRAMELEN_HEADER, &flen);

        // Validate the frame header
        if (fhead.frame_sentinel != htonl(STREAM_SENTINEL)) {
            snprintf(errstr, 1024, "TCPSSLAP Got an invalid frame sentinel, "
                     "closing connection fd %d", in_fd);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }

        fhead.frame_len = (uint32_t) ntohl(fhead.frame_len);

        // Do we have enough for the payload?
        if (rbuf->FetchLen() >= (int) (FRAMELEN_HEADER + fhead.frame_len)) {
            // Allocate a payload frame
            uint8_t *payload = new uint8_t[fhead.frame_len];

            // Mark the header as read in the ring
            rbuf->MarkRead(FRAMELEN_HEADER);

            // Fetch the payload and mark it read
            rbuf->FetchPtr(payload, fhead.frame_len, &flen);
            rbuf->MarkRead(fhead.frame_len);

            // Handle packet types
            if (fhead.frame_type == FRAMETYPE_HELO) {
                ret = ParseHelo(in_fd, (ap_helo_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_ANNOUNCE) {
                ret = ParseAnnounce(in_fd, (ap_announce_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_MESSAGE) {
                ret = ParseMessage(in_fd, (ap_message_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_KISALERT) {
                ret = ParseKisAlert(in_fd, (ap_kisalert_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_KISNETWORK) {
                ret = ParseKisNetwork(in_fd, (ap_kisnetwork_packet *) payload);
            } else {
                snprintf(errstr, 1024, "TCPSSLAP Got unknown frame type %d, "
                         "closing connection %d", fhead.frame_type, in_fd);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                ret = -1; 
            }

            delete[] payload;
        } else {
            // Drop out of while if we don't have enough to complete the packet
            break;
        }
    }

    return ret;
}

int TcpSSLApServer::ParseAnnounce(int in_fd, ap_announce_packet *annframe) {
    // printf("debug - parseannounce\n");
    tcpssl_conrec *crec = ssl_rec_map[in_fd];

    // Match the name against the certificate name and see if they're
    // kosher....
    X509_NAME *certsubj;
    char certdata[33];

    if ((certsubj = X509_get_subject_name(crec->ssl_peercert)) != NULL &&
        X509_NAME_get_text_by_NID(certsubj, NID_commonName, certdata, 33) > 0) {
        certdata[32] = '\0';
        if (strncmp((const char *) annframe->name, certdata, 32)) {
            snprintf(errstr, 1024, "TCPSSLAP Client on fd %d advertised "
                     "name '%s' doesn't match cert '%s'",
                     in_fd, (const char *) annframe->name, certdata);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }
    } else {
        snprintf(errstr, 1024, "TCPSSLAP Couldn't fetch commonName out of "
                 "peer certificate for client on fd %d claiming to have name '%s'",
                 in_fd, (const char *) annframe->name);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
   
    // Now that we have their name, see if they're rejoining an old session, or if 
    // they're new...
    for (map<int, apserver_aprec *>::iterator aitr = linked_ap_map.begin();
         aitr != linked_ap_map.end(); ++aitr) {
        if (!strncmp((const char *) annframe->name, aitr->second->name, 32)) {
            if (aitr->second->dead == 0) {
                // Bail out on an active AP that we're trying to replace...  But
                // be noisy about it
                snprintf(errstr, 1024, "TCPSSLAP AP on fd %d advertised "
                         "name '%s' collides with active AP id %u",
                         in_fd, aitr->second->name, aitr->second->station_id);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return -1;
            } else {
                // See if the AP has been configured and if the info about it
                // matches... this is a damn ugly big if but it doesn't get
                // executed too often...
                apserver_aprec *trec = aitr->second;
                if (trec->configured == 0 || 
                    annframe->zone != trec->zone ||
                    mac_addr(annframe->bssid0) != trec->interfaces[0].mac ||
                    mac_addr(annframe->bssid1) != trec->interfaces[1].mac ||
                    mac_addr(annframe->bssid2) != trec->interfaces[2].mac ||
                    mac_addr(annframe->bssid3) != trec->interfaces[3].mac ||
                    ((ntohl(annframe->cf_bitmap0) & AP_IFFLAG_AVAILABLE) != 
                     (trec->interfaces[0].cf_bitmap & AP_IFFLAG_AVAILABLE)) ||
                    ((ntohl(annframe->cf_bitmap1) & AP_IFFLAG_AVAILABLE) != 
                     (trec->interfaces[1].cf_bitmap & AP_IFFLAG_AVAILABLE)) ||
                    ((ntohl(annframe->cf_bitmap2) & AP_IFFLAG_AVAILABLE) != 
                     (trec->interfaces[2].cf_bitmap & AP_IFFLAG_AVAILABLE)) ||
                    ((ntohl(annframe->cf_bitmap3) & AP_IFFLAG_AVAILABLE) != 
                     (trec->interfaces[3].cf_bitmap & AP_IFFLAG_AVAILABLE))) {
                    // Destroy it if it doesn't match and we need to do a
                    // full reconfigure...
                    snprintf(errstr, 1024, "TCPSSLAP AP on fd %d looks like %u '%s' "
                             "but is unconfigured or doesn't match previous config.",
                             in_fd, trec->station_id, trec->name);
                    msgbus->InjectMessage(errstr, MSGFLAG_INFO);
                    DestroyRecords(trec);
                } else {

                    // Configure and activate the ap immediately
                    snprintf(errstr, 1024, "TCPSSLAP Reinserting '%s' into AP slot %d",
                             trec->name, trec->station_id);
                    msgbus->InjectMessage(errstr, MSGFLAG_INFO);

                    // Move the fd record
                    linked_ap_map.erase(aitr);
                    linked_ap_map[in_fd] = trec;
                    
                    trec->fd = in_fd;
                    gettimeofday(&(trec->last_helo), NULL);
                    trec->dead = 0;
                    trec->last_packet_local = time(0);

                    // Inherit their sequence ID
                    trec->rx_seqid = (uint64_t) kis_ntoh64(annframe->sequence_id);
                   
                    // Zero the byte counters
                    for (unsigned int i = 0; i < 4; i++) {
                        trec->interfaces[i].tx_bytecount = 0;
                        trec->interfaces[i].rx_bytecount = 0;
                        trec->interfaces[i].tx_packetcount = 0;
                        trec->interfaces[i].rx_packetcount = 0;
                        trec->interfaces[i].tx_oldbcount = 0;
                        trec->interfaces[i].rx_oldbcount = 0;
                    }
                    
                    ConfigureAp(trec);
                    ActivateAp(trec);

                    // We're done, get out
                    return 0;
                }
                // end active
            }
            // end match
        }
        // end for
    }

    
    apserver_aprec *arec = new apserver_aprec;
    memset(arec, 0, sizeof(apserver_aprec));

    strncpy(arec->name, (const char *) annframe->name, 32);

    // Set the indexes
    for (unsigned int x = 0; x < 4; x++) {
        arec->interfaces[x].index = x;
        arec->interfaces[x].owning_ap = arec;
        arec->interfaces[x].tx_bytecount = 0;
        arec->interfaces[x].rx_bytecount = 0;
        arec->interfaces[x].tx_packetcount = 0;
        arec->interfaces[x].rx_packetcount = 0;
        arec->interfaces[x].tx_oldbcount = 0;
        arec->interfaces[x].rx_oldbcount = 0;
    }

    // Copy the MAC over
    arec->interfaces[0].mac = mac_addr(annframe->bssid0);
    arec->interfaces[1].mac = mac_addr(annframe->bssid1);
    arec->interfaces[2].mac = mac_addr(annframe->bssid2);
    arec->interfaces[3].mac = mac_addr(annframe->bssid3);
    // Copy it to their "innate" mac before we change things, too
    arec->interfaces[0].innate_mac = arec->interfaces[0].mac;
    arec->interfaces[1].innate_mac = arec->interfaces[1].mac;
    arec->interfaces[2].innate_mac = arec->interfaces[2].mac;
    arec->interfaces[3].innate_mac = arec->interfaces[3].mac;

    arec->interfaces[0].cf_bitmap = ntohl(annframe->cf_bitmap0);
    arec->interfaces[1].cf_bitmap = ntohl(annframe->cf_bitmap1);
    arec->interfaces[2].cf_bitmap = ntohl(annframe->cf_bitmap2);
    arec->interfaces[3].cf_bitmap = ntohl(annframe->cf_bitmap3);
    
    arec->zone = annframe->zone;
    arec->station_id = next_station_id++;
    arec->fd = in_fd;
    arec->rx_seqid = (uint64_t) kis_ntoh64(annframe->sequence_id);

    // Fill in the helo time
    gettimeofday(&(arec->last_helo), NULL);

    arec->last_packet_local = time(0);

    linked_ap_map[in_fd] = arec;

    // Fill in the latest time
    arec->update_time = time(0);

    // Mark their zone as dirty
    zone_map[arec->zone].zone_dirty = 1;

    zone_map[arec->zone].zone_id = arec->zone;
    zone_map[arec->zone].update_time = time(0);
    zone_map[arec->zone].num_alerts = 0;
    zone_map[arec->zone].num_attacks = 0;
    
    snprintf(errstr, 1024, "TCPSSLAP Added new AP id %u name '%s' zone %u interfaces "
             "%s %s %s %s", arec->station_id, arec->name, arec->zone,
             arec->interfaces[0].mac.Mac2String().c_str(),
             arec->interfaces[1].mac.Mac2String().c_str(),
             arec->interfaces[2].mac.Mac2String().c_str(),
             arec->interfaces[3].mac.Mac2String().c_str());
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);
   
    // Generate a timer event to configure all aps if one isn't queued already
    if (config_pending == 0) {
        timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * config_delay,
                                   NULL, 0, &SSLApServerConfigureTimer, 
                                   (void *) this);
        config_pending = 1;
    }

    return 1;
}

TcpSSLApServer::apserver_aprec 
    *TcpSSLApServer::ParseCommon(int in_fd, ap_common_payload *comframe) {
    // Look for a record for them
    apserver_aprec *arec;
    map<int, apserver_aprec *>::iterator aitr = linked_ap_map.find(in_fd);
    if (aitr == linked_ap_map.end()) {
        snprintf(errstr, 1024, "TCPSSLAP Data recieved on fd %d from unknown station "
                 "id %d", in_fd, ntohs(comframe->station_id));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return NULL;
    }
    arec = aitr->second;
    
    // Protect against replay attacks
    unsigned long int seqid = (long int) kis_ntoh64(comframe->sequence_id);
    if (seqid < arec->rx_seqid || 
        seqid > arec->rx_seqid + STREAM_SEQ_WINDOW) {
        snprintf(errstr, 1024, "TCPSSLAP Data recieved on fd %d out-of-sequence for "
                 "station id %d got %lu expected %lu", in_fd, arec->station_id,
                 (unsigned long int) seqid, 
                 (unsigned long int) arec->rx_seqid - 1);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return NULL;
    }

    // Update the sequence in the main records
    arec->rx_seqid = seqid;

    // Update the time
    arec->last_packet_local = time(0);
    
    return arec;
}

int TcpSSLApServer::ParseHelo(int in_fd, ap_helo_packet *heloframe) {
    // printf("debug - parsehelo\n");
    apserver_aprec *arec;
    if ((arec = ParseCommon(in_fd, (ap_common_payload *) heloframe)) == NULL)
        return -1;
   
    // Do smart things with the frame
    arec->last_helo.tv_sec = ntohl(heloframe->station_timestamp_sec);
    arec->last_helo.tv_usec = ntohl(heloframe->station_timestamp_usec);

    arec->load_avg_major = heloframe->load_avg_major;
    arec->load_avg_minor = heloframe->load_avg_minor;
   
    arec->interfaces[0].clientcount = ntohl(heloframe->clientcount0);
    arec->interfaces[1].clientcount = ntohl(heloframe->clientcount1);
    arec->interfaces[2].clientcount = ntohl(heloframe->clientcount2);
    arec->interfaces[3].clientcount = ntohl(heloframe->clientcount3);

    arec->interfaces[0].framecount = kis_ntoh64(heloframe->framecount0);
    arec->interfaces[1].framecount = kis_ntoh64(heloframe->framecount1);
    arec->interfaces[2].framecount = kis_ntoh64(heloframe->framecount2);
    arec->interfaces[3].framecount = kis_ntoh64(heloframe->framecount3);
    
    arec->interfaces[0].errorcount = kis_ntoh64(heloframe->errorcount0);
    arec->interfaces[1].errorcount = kis_ntoh64(heloframe->errorcount1);
    arec->interfaces[2].errorcount = kis_ntoh64(heloframe->errorcount2);
    arec->interfaces[3].errorcount = kis_ntoh64(heloframe->errorcount3);

    arec->interfaces[0].tx_packetcount = kis_ntoh64(heloframe->txpacketcount0);
    arec->interfaces[1].tx_packetcount = kis_ntoh64(heloframe->txpacketcount1);
    arec->interfaces[2].tx_packetcount = kis_ntoh64(heloframe->txpacketcount2);
    arec->interfaces[3].tx_packetcount = kis_ntoh64(heloframe->txpacketcount3);

    arec->interfaces[0].rx_packetcount = kis_ntoh64(heloframe->rxpacketcount0);
    arec->interfaces[1].rx_packetcount = kis_ntoh64(heloframe->rxpacketcount1);
    arec->interfaces[2].rx_packetcount = kis_ntoh64(heloframe->rxpacketcount2);
    arec->interfaces[3].rx_packetcount = kis_ntoh64(heloframe->rxpacketcount3);

    uint64_t nbtc[4];
    uint64_t nbrc[4];
    nbtc[0] = kis_ntoh64(heloframe->txbytecount0);
    nbtc[1] = kis_ntoh64(heloframe->txbytecount1);
    nbtc[2] = kis_ntoh64(heloframe->txbytecount2);
    nbtc[3] = kis_ntoh64(heloframe->txbytecount3);

    nbrc[0] = kis_ntoh64(heloframe->rxbytecount0);
    nbrc[1] = kis_ntoh64(heloframe->rxbytecount1);
    nbrc[2] = kis_ntoh64(heloframe->rxbytecount2);
    nbrc[3] = kis_ntoh64(heloframe->rxbytecount3);

    for (unsigned int x = 0; x < 4; x++) {
        if (arec->interfaces[x].tx_oldbcount < nbtc[x]) {
            // If we've wrapped around, add the delta to max and then add what
            // we've wrapped to.
            arec->interfaces[x].tx_bytecount += 
                ((unsigned long long int) ULONG_MAX - arec->interfaces[x].tx_oldbcount) + 
                nbtc[x];
        } else {
            // Otherwise add the delta
            arec->interfaces[x].tx_bytecount += nbtc[x] - arec->interfaces[x].tx_oldbcount;
        }

        if (arec->interfaces[x].rx_oldbcount < nbrc[x]) {
            // If we've wrapped around, add the delta to max and then add what
            // we've wrapped to.
            arec->interfaces[x].rx_bytecount += 
                ((unsigned long long int) ULONG_MAX - arec->interfaces[x].rx_oldbcount) + 
                nbrc[x];
        } else {
            // Otherwise add the delta
            arec->interfaces[x].rx_bytecount += nbrc[x] - arec->interfaces[x].rx_oldbcount;
        }
    }
    

    for (unsigned int x = 0; x < 4; x++) {
        arec->interfaces[x].old_cf_bitmap = arec->interfaces[x].cf_bitmap;
        // Update the time
        arec->interfaces[x].update_time = time(0);
    }

    arec->interfaces[0].cf_bitmap = ntohl(heloframe->cf_bitmap0);
    arec->interfaces[1].cf_bitmap = ntohl(heloframe->cf_bitmap1);
    arec->interfaces[2].cf_bitmap = ntohl(heloframe->cf_bitmap2);
    arec->interfaces[3].cf_bitmap = ntohl(heloframe->cf_bitmap3);

    // Spew about interfaces failing
    for (unsigned int x = 0; x < 4; x++) {
        if ((arec->interfaces[x].cf_bitmap & AP_IFFLAG_ERROR) &&
            (arec->interfaces[x].old_cf_bitmap & AP_IFFLAG_ERROR) == 0) {
            snprintf(errstr, 1024, "Station '%s' in zone %u reports interface %u "
                     "('%s') is in error and is no longer available.",
                     arec->name, arec->zone, x, 
                     arec->interfaces[x].name);

            zone_map[arec->zone].zone_dirty = 1;
            if (config_pending == 0) {
                timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * config_delay,
                                           NULL, 0, &SSLApServerConfigureTimer, 
                                           (void *) this);
                config_pending = 1;
            }
        } else if ((arec->interfaces[x].cf_bitmap & AP_IFFLAG_ERROR) == 0 &&
                   (arec->interfaces[x].old_cf_bitmap & AP_IFFLAG_ERROR)) {
            snprintf(errstr, 1024, "Station '%s' in zone %u reports interface %u "
                     "('%s') is no longer in error is available.",
                     arec->name, arec->zone, x, 
                     arec->interfaces[x].name);

            zone_map[arec->zone].zone_dirty = 1;
            if (config_pending == 0) {
                timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * config_delay,
                                           NULL, 0, &SSLApServerConfigureTimer, 
                                           (void *) this);
                config_pending = 1;
            }
        }
    }
    
    /* 
       printf("debug - Got HELO frame from %u %hu.%hu %u %lu/%lu %u %lu/%lu "
           "%u %lu/%lu %u %lu/%lu\n",
           arec->station_id, 
           arec->load_avg_major, arec->load_avg_minor,
           arec->interfaces[0].clientcount,
           arec->interfaces[0].framecount,
           arec->interfaces[0].errorcount,
           arec->interfaces[1].clientcount,
           arec->interfaces[1].framecount,
           arec->interfaces[1].errorcount,
           arec->interfaces[2].clientcount,
           arec->interfaces[2].framecount,
           arec->interfaces[2].errorcount,
           arec->interfaces[3].clientcount,
           arec->interfaces[3].framecount,
           arec->interfaces[3].errorcount);
    */

    return 1;
}

int TcpSSLApServer::ParseMessage(int in_fd, ap_message_packet *msgframe) {
    // printf("debug - parsemessage\n");
    apserver_aprec *arec;
    if ((arec = ParseCommon(in_fd, (ap_common_payload *) msgframe)) == NULL)
        return -1;
   
    int msgflags = ntohl(msgframe->msgflags);
    char msg[1024];

    // Translate the flags
    int mesbus_flags = 0;
    if (msgflags & AP_MSGFLAG_DEBUG)
        mesbus_flags |= MSGFLAG_DEBUG;
    if (msgflags & AP_MSGFLAG_INFO)
        mesbus_flags |= MSGFLAG_INFO;
    if (msgflags & AP_MSGFLAG_ERROR)
        mesbus_flags |= MSGFLAG_ERROR;
    // We dont' let a fatal propogate up
    if (msgflags & AP_MSGFLAG_FATAL)
        mesbus_flags |= MSGFLAG_ERROR;
   
    snprintf(msg, 1024, "TCPSSLAP Message from '%s': %s", arec->name,
             (char *) &(msgframe->message));
    msgbus->InjectMessage(msg, mesbus_flags);

    return 1;
}

int TcpSSLApServer::ParseKisAlert(int in_fd, ap_kisalert_packet *alertframe) {
    // printf("debug - parsekisalert\n");
    apserver_aprec *arec;
    if ((arec = ParseCommon(in_fd, (ap_common_payload *) alertframe)) == NULL)
        return -1;

    TcpKismetClient::alert_rec *alrrec = new TcpKismetClient::alert_rec;
    
    alrrec->type = (TcpKismetClient::kis_alert_type) alertframe->type;
    alrrec->bssid_mac = alertframe->bssid_mac;
    alrrec->source_mac = alertframe->source_mac;
    alrrec->dest_mac = alertframe->dest_mac;
    alrrec->other_mac = alertframe->other_mac;
    alrrec->timestamp = (time_t) ntohl(alertframe->timestamp);
    alrrec->channel = ntohs(alertframe->channel);
   
    alrrec->tracker_time = time(0);
    alrrec->ext = arec;

    if (seen_alert_vec.size() >= 20) {
        delete seen_alert_vec[0];
        seen_alert_vec.erase(seen_alert_vec.begin());
    }

    alerts_dirty = 1;

    seen_alert_vec.push_back(alrrec);
    
    return 1;
}

int TcpSSLApServer::ParseKisNetwork(int in_fd, ap_kisnetwork_packet *netframe) {
    // printf("debug - parsekisnetwork\n");
    apserver_aprec *arec;
    if ((arec = ParseCommon(in_fd, (ap_common_payload *) netframe)) == NULL)
        return -1;

    TcpKismetClient::network_rec *nrec;
    mac_addr netbssid = netframe->bssid_mac;

    macmap<TcpKismetClient::network_rec *>::iterator nmi = 
        seen_network_map.find(netbssid);

    if (nmi == seen_network_map.end()) {
        nrec = new TcpKismetClient::network_rec;
        nrec->bssid_mac = netbssid;
        seen_network_map.insert(netbssid, nrec);
    } else {
        nrec = *(nmi->second);
    }

    nrec->type = (TcpKismetClient::kis_network_type) netframe->type;
    nrec->ssid = (char *) netframe->ssid;
    nrec->attributes = (uint32_t) ntohl(netframe->attributes);
    nrec->channel = ntohs(netframe->channel);
    nrec->first_time = (time_t) ntohl(netframe->first_time);
    nrec->last_time = (time_t) ntohl(netframe->last_time);
   
    nrec->tracker_time = time(0);
    nrec->ext = arec;

    // printf("debug - %d - got update on network %s\n", nrec->tracker_time, nrec->bssid_mac.Mac2String().c_str());

    networks_dirty = 1;

    return 1;
}

int TcpSSLApServer::QueuePacket(int in_fd, apserver_aprec *in_arec,
                                stream_frame *in_frame) {
    if (in_arec->dead) {
        snprintf(errstr, 1024, "Attempt to write to dead client %u, they have "
                 "no ring buffer, this is bad.", in_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return 0;
    }

    // printf("debug - queuing packet type %d\n", in_frame->frame_type);
    int hostlen = ntohl(in_frame->frame_len);
    RingBuffer *write_buf = write_buf_map[in_fd];

    // Update the sequence ID
    mn_common_payload *cpay = (mn_common_payload *) in_frame->payload;
    cpay->sequence_id = kis_hton64((uint64_t) in_arec->tx_seqid++);
 
    // Queue it into the write buffer
    if (write_buf->InsertDummy(FRAMELEN_HEADER + hostlen) == 0) {
        snprintf(errstr, 1024, "TCPSSLAP no room in write ring buffer to insert "
                 "queued packet.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    write_buf->InsertData((uint8_t *) in_frame, FRAMELEN_HEADER);
    write_buf->InsertData((uint8_t *) in_frame->payload, hostlen);

    return FRAMELEN_HEADER + hostlen;
}

stream_frame *TcpSSLApServer::Banrec2Frame(apserver_banrec *in_brec) {
    // Queue up an accept packet for them with their id
    stream_frame *retframe = new stream_frame;
    mn_banuser_packet *banpayload = new mn_banuser_packet;

    retframe->frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    retframe->frame_type = FRAMETYPE_BANUSER;
    retframe->frame_len = (uint32_t) htonl(FRAMELEN_BANUSER);
    retframe->payload = (uint8_t *) banpayload;

    // Copy the banned MAC addr
    for (unsigned int x = 0; x < 6; x++)
        banpayload->banmac[x] = in_brec->bannedmac[x];

    banpayload->ban_time = (uint32_t) htonl(in_brec->ban_time);
    banpayload->expire_time = (uint32_t) htonl(in_brec->expire_time);
   
    return retframe;
}

void TcpSSLApServer::ProcessActiveEvents() {
    // printf("debug - processactiveevents\n");
    if (alerts_dirty) {
        for (int x = (int) (seen_alert_vec.size() - 1); x >= 0; --x) {
            // Drop out if they're no longer new
            if (seen_alert_vec[x]->tracker_time < last_alert_time)
                break;

            // Alert events that cause us to try to change MAC addresses
            if (seen_alert_vec[x]->type == TcpKismetClient::kis_alrt_bsstimestamp ||
                seen_alert_vec[x]->type == TcpKismetClient::kis_alrt_deauthflood ||
                seen_alert_vec[x]->type == TcpKismetClient::kis_alrt_bcastdiscon) {
                // Find the interface that owns the BSSID this is affecting
                // apserver_aprec *owner_ap;
                apserver_apinterface *owner_if;

                macmap<TcpSSLApServer::apserver_apinterface *>::iterator apitr =
                    bssid_if_map.find(seen_alert_vec[x]->bssid_mac);
                if (apitr == bssid_if_map.end()) {
                    /*
                    printf("debug - alert for %s which we don't own\n",
                           seen_alert_vec[x]->bssid_mac.Mac2String().c_str());
                    */
                    continue;
                }

                owner_if = *(apitr->second);

                // Log the error in our stats
                zone_map[owner_if->owning_ap->zone].num_alerts++;

                // Don't process alerts for networks that aren't APs
                if ((owner_if->cf_bitmap & AP_ACTIVEAP) != AP_ACTIVEAP)
                    continue;
                
                // Short out the rest of this since it doesn't work right now
                snprintf(errstr, 1024, "BSSID %s under attack from "
                         "BSSSPOOF/DEAUTHFLOOD/DISCONFLOOD expect degraded performance "
                         "for users on this node.", 
                         seen_alert_vec[x]->bssid_mac.Mac2String().c_str());
                msgbus->InjectMessage(errstr, MSGFLAG_INFO);
                continue;
               
                // WE NEVER GET HERE
#if 0
                owner_if = *(apitr->second);
                owner_ap = owner_if->owning_ap;

                // Have we changed BSSID in the last minute?  If so, we don't want to
                // thash and keep changing it, so do nothing
                if ((time(0) - owner_if->last_mac_change) < 60) {
                    snprintf(errstr, 1024, "BSSID %s under attack, but will not shift "
                             "BSSID within 60 seconds of a previous shift",
                             owner_if->mac.Mac2String().c_str());
                    continue;
                }

                // Set the new MAC and clear out the old records
                bssid_if_map.erase(owner_if->mac);
                mac_addr old_mac = owner_if->mac;
                owner_if->mac = next_bssid++;
                bssid_if_map.insert(owner_if->mac, owner_if);
                owner_if->last_mac_change = time(0);
                PushBSSID(owner_ap, owner_if->index);

                snprintf(errstr, 1024, "Shifted BSSID of %s to %s to avoid an attack",
                         old_mac.Mac2String().c_str(), owner_if->mac.Mac2String().c_str());
                msgbus->InjectMessage(errstr, MSGFLAG_INFO);

                // Attempt to dispatch a smite for spoofed APs we moved
                if (seen_alert_vec[x]->type == TcpKismetClient::kis_alrt_bsstimestamp) 
                    SmiteAP(owner_ap, old_mac, seen_alert_vec[x]->channel);
                
#endif
            }

        }
       
        last_alert_time = time(0);
        alerts_dirty = 0;
    }

    // Only do spoof smites once a minute
    if (networks_dirty && (time(0) - last_ssid_smiting) > (int) smite_delay) {
        for (macmap<TcpKismetClient::network_rec *>::iterator mi = seen_network_map.begin();
             mi != seen_network_map.end(); ++mi) {
            TcpKismetClient::network_rec *nrec = *(mi->second);

            // Disregard unchanged ones
            if (nrec->tracker_time < last_network_time)
                continue;

            // Don't pay attention to the network if we've smited it within the ignore time
            macmap<time_t>::iterator itr = rawtx_time_map.find(nrec->bssid_mac);
            if (itr != rawtx_time_map.end() &&
                (time(0) - *(itr->second)) < (int) smite_ignore)
                continue;

            // Look for spoofed ones, require WEP be turned on to count them as spoofed
            if (!strncmp(nrec->ssid.c_str(), base_ssid, 32)) {
                // We know about them, continue
                if (bssid_if_map.find(nrec->bssid_mac) != bssid_if_map.end()) {
                    continue;
                }
               
                if ((nrec->attributes & KISNETATTRIB_WEP) != 0) {
                    continue;
                }

                SmiteAP((apserver_aprec *) nrec->ext, nrec->bssid_mac, 
                        nrec->channel);

                last_ssid_smiting = time(0);

                rawtx_time_map.insert(nrec->bssid_mac, time(0));
            }
        }

        last_network_time = time(0);
        networks_dirty = 0;
    }
}

void TcpSSLApServer::SmiteAP(apserver_aprec *in_sarec, mac_addr in_bssid, int in_channel) {
    // printf("debug - smite\n");
    // Find an interface to transmit on inside the zone
    apserver_apinterface *transmitter = NULL;

    for (macmap<TcpSSLApServer::apserver_apinterface *>::iterator x = bssid_if_map.begin();
         x != bssid_if_map.end(); ++x) {
        apserver_apinterface *ifc = *(x->second);

        // Filter out-of-zone, inappropriate types or busy APs
        if (ifc->owning_ap->zone != in_sarec->zone)
            continue;

        if ((ifc->cf_bitmap & AP_IFFLAG_RAWTX) == 0)
            continue;

        if ((ifc->cf_bitmap & AP_IFFLAG_BUSY))
            continue;

        // Fallthrough to a valid rawtransmitter
        transmitter = ifc;
        break;
    }

    if (transmitter == NULL) {
        snprintf(errstr, 1024, "Would have performed a rawtx AP smite on "
                 "%s channel %d but no rawtx devices were available.",
                 in_bssid.Mac2String().c_str(), in_channel);
        msgbus->InjectMessage(errstr, MSGFLAG_INFO);
        return;
    }

    snprintf(errstr, 1024, "Sending rawtx AP smite to %s channel %d ",
            in_bssid.Mac2String().c_str(), in_channel);
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);

    // Update our stats
    zone_map[in_sarec->zone].num_attacks++;
    
    // Push a config down to the AP
    stream_frame retframe;
    mn_smite_packet smitepay;

    retframe.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    retframe.frame_type = FRAMETYPE_SMITE;
    retframe.frame_len = (uint32_t) htonl(FRAMELEN_SMITE);
    retframe.payload = (uint8_t *) &smitepay;

    smitepay.if_index = (uint8_t) (transmitter->index & 0xFF);

    for (unsigned int x = 0; x < 6; x++) 
        smitepay.bssid[x] = in_bssid[x];

    smitepay.channel = (uint16_t) htons(in_channel);

    // Queue it out
    QueuePacket(transmitter->owning_ap->fd, 
                transmitter->owning_ap, &retframe);

    // Set the local busy flag to protect us from doubling up before the 
    // HELO frame comes in when we're doing things
    transmitter->cf_bitmap |= AP_IFFLAG_BUSY;
}

