/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TCPSSLAPSERVER_H__
#define __TCPSSLAPSERVER_H__

#include "macaddr.h"
#include "tcpsslserver.h"
#include "ap_stream_proto.h"
#include "tcpkismetclient.h"

// Finally drill down through all the abstraction and do something useful with 
// ourselves

// C-space event handlers to catch timed events and trigger calls into the function
// itself
// Timer pop to configure all currently known APs
int SSLApServerConfigureTimer(Timetracker::timer_event *evt, void *parm);
// Timer pop to process kismet events and dispatch active responses
int SSLApServerActiveEventTimer(Timetracker::timer_event *evt, void *parm);

// Forward definition of the monitor class so we can be friends with it
class TcpMonitorServer;

class TcpSSLApServer : public TcpSSLServer {
public:
    // Stuff we need to track about an interface in each AP
    typedef struct apserver_aprec;

    typedef struct apserver_apinterface {
        // Name of the interface
        char name[32];
        // Index in the AP
        int index;
        apserver_aprec *owning_ap;
        // MAC - this is bad to memset to 0 like we do but we create it with a 
        // mac_addr() constructor so it works out in the end, just remember we're
        // dumb here and don't shoot ourselves in the foot later
        mac_addr mac;
        mac_addr innate_mac;
        // Channel we assigned them
        unsigned int channel;
        // Config bitmap to control what this interface is
        unsigned int cf_bitmap;
        unsigned int old_cf_bitmap;
        // Are we configured this conf-iteration?
        unsigned int conf_iter_set;

        // Last time we changed the BSSID on this interface so we don't
        // change it too often
        time_t last_mac_change;

        // Info from the AP
        // nclients
        int clientcount;
        // nframes
        unsigned long long int framecount;
        // nerrors
        unsigned long long int errorcount;

        // xfer stats
        unsigned long long int tx_packetcount;
        unsigned long long int tx_bytecount;
        unsigned long long int rx_packetcount;
        unsigned long long int rx_bytecount;

        unsigned long long int tx_oldbcount;
        unsigned long long int rx_oldbcount;

        // Bits for the monitor hooks
        time_t update_time;
    };

    // Stuff we need to track about each AP talking to us
    typedef struct apserver_aprec {
        // Is it locally configured?
        int localconf;
        // Has the config been pushed out?
        int configured;
        // Have we activated it?
        int active;
        // Is the AP dead and disconnected?
        int dead;

        unsigned int station_id;
        char name[32];
        unsigned int zone;

        // Fd for looking it up in other maps
        int fd;

        // Last time on the remote system
        struct timeval last_helo;

        // Last time we saw something on the local system
        time_t last_packet_local;

        // Last seen sequence
        unsigned long int rx_seqid;
        // Last sent sequence
        unsigned long int tx_seqid;

        // Data FROM the AP
        uint8_t load_avg_major;
        uint8_t load_avg_minor;

        // Interfaces
        apserver_apinterface interfaces[4];

        // Bits for the monitor hooks
        time_t update_time;
    };

    // Track banned MAC addresses for users who abuse once
    // associated
    typedef struct apserver_banrec {
        mac_addr bannedmac;
        time_t ban_time;
        time_t expire_time;
    };

    // Used for building and tracking zones
    typedef struct apserver_zonerec {
        int zone_id;
        // Track stuff
        vector<apserver_aprec *> aps;
        vector<apserver_apinterface *> interfaces;
        // Zone needs to be updated
        int zone_dirty;
        // Number of interfaces on each task
        int num_sniffer;
        int num_rawtx;
        int num_ap;
        // Number of alerts seen for this zone
        int num_alerts;
        // Number of attacks we've initiated
        int num_attacks;
        // Bits for the monitor hooks
        time_t update_time;
    };

    TcpSSLApServer();

    virtual void KillConnection(int in_fd);

    // Set the time limits from the config file
    virtual void SetTimeLimits(unsigned int in_conf, unsigned int in_reconf, 
                               unsigned int in_idle, unsigned int in_smitedelay,
                               unsigned int in_smiteignore);
    
    // Set the allocation metrics
    virtual void SetInterfaceAllocation(unsigned int in_sniffers, 
                                        unsigned int in_rawtx,
                                        mac_addr in_base_bssid,
                                        const char *in_base_ssid);

    virtual void Tick();

protected:
    // AP activation stuff triggered by timers
    // Configure all APs that have been announced to us
    virtual void ConfigureAllAps();

    // Send configure/reconfigure packets to the actual AP based on the
    // in-server config record, or smite a fake AP
    virtual void ConfigureAp(apserver_aprec *in_arec);
    virtual void PushBSSID(apserver_aprec *in_arec, int in_index);
    virtual void ActivateAp(apserver_aprec *in_arec);
    virtual void SmiteAP(apserver_aprec *in_sarec, mac_addr in_bssid, int in_channel);

    // Is a configure pending?  If so, don't register one with the timetracker
    int config_pending;

    // Destroy the records fully
    virtual void DestroyRecords(apserver_aprec *in_arec);
    
    // high-level packet parse
    virtual int ParseData(int in_fd);

    // Process individual packet types
    // Announce frames create an entry that an AP is coming online
    virtual int ParseAnnounce(int in_fd, ap_announce_packet *annframe);

    // Common verification for all but announce frames
    virtual TcpSSLApServer::apserver_aprec *ParseCommon(int in_fd, 
                                                        ap_common_payload *commonframe);
    // Helo frames happen at regular intervals
    virtual int ParseHelo(int in_fd, ap_helo_packet *heloframe);

    // Text messages
    virtual int ParseMessage(int in_fd, ap_message_packet *msgframe);

    // Kismet bits - alerts and network info so we can build our own list of networks
    // and report how busy things are and if we're being spoofed
    virtual int ParseKisAlert(int in_fd, ap_kisalert_packet *alertframe);
    virtual int ParseKisNetwork(int in_fd, ap_kisnetwork_packet *netframe);

    // Queue a packet into a given clients write buffer
    virtual int QueuePacket(int in_fd, apserver_aprec *in_arec,
                            stream_frame *in_frame);

    // Dispatch active events
    virtual void ProcessActiveEvents();

    // Turn a banrec into a banuser packet, because I hate retyping
    // code thats needed in a few places
    stream_frame *Banrec2Frame(apserver_banrec *in_brec);

    // ID of the active events timer
    int active_id;
    
    // Time delays before triggers happen
    // Time from first connect to mass config
    unsigned int config_delay;
    // Time from an AP disconnecting until the netwide reconfiguration
    unsigned int reconfigure_delay;
    // Time for an AP to be unresponsive before we consider it dead
    unsigned int deadidle_delay;
    // Time between smitings
    unsigned int smite_delay;
    // Time to ignore an AP after smiting
    unsigned int smite_ignore;

    // Allocations per zone
    unsigned int sniffers_per_zone;
    unsigned int rawtx_per_zone;
    
    // FD to AP record
    map<int, TcpSSLApServer::apserver_aprec *> linked_ap_map;
    // Secondary mapping to link a BSSID to a given AP so we can
    // see if we're being spoofed and that sort of good stuff
    macmap<TcpSSLApServer::apserver_apinterface *> bssid_if_map;
    // Map zone numbers to if we know an AP in that zone has been
    // changed.   This is sort of a kluge.
    map<int, apserver_zonerec> zone_map;
    // Time we initiated a rawtx, so we know not to look at kismets 
    // detecting this network for now.  
    macmap<time_t> rawtx_time_map;

    // Next ID to assign a joining station
    unsigned int next_station_id;

    // Next BSSID to allocate
    mac_addr next_bssid;

    // SSID to set
    char base_ssid[32];

    // MAC addresses of users we've decided to ban
    vector<TcpSSLApServer::apserver_banrec *> banned_mac_vec;

    // Map of networks that have been seen by our sniffer drones
    macmap<TcpKismetClient::network_rec *> seen_network_map;
    // Last network timestamp we processed
    time_t last_network_time;
    // Networks dirty
    int networks_dirty;
    
    // List of alerts generated by our sniffer drones
    vector<TcpKismetClient::alert_rec *> seen_alert_vec;
    // Last alert timestamp we looked at
    time_t last_alert_time;
    // Alerts dirty
    int alerts_dirty;

    // Last time we smote someone
    time_t last_ssid_smiting;

    // Lets be pals with the timer events
    friend int SSLApServerConfigureTimer(Timetracker::timer_event *evt, void *parm);
    friend int SSLApServerActiveEventTimer(Timetracker::timer_event *evt, void *parm);

    // Let the monitor server hook into us to extract info
    friend class TcpMonitorServer;
};

#endif

